
{-|
  Module      :  Misc

  Maintainer  :  Misha Aizatulin <avatar@hot.ee>
  Portability :  portable

  General purpose utilities of all kind.
-} 

module Lib.Misc 
(
-- * Lists
  getIndex,
  getRange,
  replaceIndex,
  updateIndex,
  replaceRange,
  pad,
  safeInit,
  hasLen,

-- * Trees
  NodePos,
  zipTree,
  numberTree,
  findTreeNode,
  inTree,
  updateSubTree,
  updateRootLabel,
  updateSubForest,
  getTreeNode,
  nodePos2Number,

-- * String
  trim,
  tokenize,
  runReadP
)
where

import Data.Tree as Tree

import Control.Monad (msum)

import Data.Typeable

-- for tokenize: 
import Text.Read.Lex
import Text.ParserCombinators.ReadP hiding (get)
import Data.Char (isSpace)

-- for numberTree
import Control.Monad.State

import Prelude hiding (lex)

-- ==============================
-- * Lists
-- ==============================

getIndex :: [a] -> Int -> Maybe a
getIndex (x:xs) 0 = Just x
getIndex (x:xs) i = getIndex xs (i - 1)
getIndex []   _ = Nothing

-- | Assumes that the first int is <= the last
getRange :: [a] -> (Int, Int) -> Maybe [a]
getRange xs (i, j) =
  if xs `hasAtLeast` (j + 1) then Just $ take (j - i + 1) $ drop i xs
  else Nothing

{-| 
  Assumes the index is correct as applied to this list.
  For error check use 'getIndex'.
-}
replaceIndex :: Int -> a -> [a] -> [a]
replaceIndex i x = updateIndex i (const x)

{-| 
  Same as 'replaceIndex' but takes an update function instead of a new element.
-}
updateIndex :: Int -> (a -> a) -> [a] -> [a]
updateIndex i f xs = xs1 ++ (f x):xs2 where
  (xs1, x:xs2) = splitAt i xs

{-| 
  Assumes the range is correct as applied to this list.
  For error check use 'getRange'
-}
replaceRange :: (Int, Int) -> [a] {-^ replacement -} -> [a] {-^ list to replace in -} -> [a]
replaceRange (i, j)  xs' xs = xs1 ++ xs' ++ xs2 where
  (xs1, xs2') = splitAt i xs
  xs2 = drop (j - i + 1) xs2'

hasLen :: [a] -> Int -> Bool
hasLen [] 0  = True
hasLen (x:xs) i | i > 0 = hasLen xs (i - 1)
hasLen _ _  = False

hasAtLeast :: [a] -> Int -> Bool
hasAtLeast xs 0 = True
hasAtLeast (x:xs) i | i > 0 = hasAtLeast xs (i - 1)
hasAtLeast _ _  = False

pad :: a   {-^ element to pad with -} 
    -> Int {-^ desired length -} 
    -> [a] {-^ list to pad -}
    -> [a] {-^ padded list -}
pad a i as = as ++ (replicate (max 0 (i - length as)) a)

{-|
  'id' on empty lists and 'init' on non-empty lists.
-}
safeInit :: [a] -> [a]
safeInit [] = []
safeInit xs = init xs

-- ==============================
-- * Trees
-- ==============================

-- TODO: think of using some generics

-- | Address of a node in the tree, read from left to right.
type NodePos = [Int]

findTreeNode :: (a -> Bool) -> Tree a -> Maybe NodePos
findTreeNode f = find [] where
  find pos node =
    msum $ here:(zipWith find [pos++[i] | i <- [0..]] (subForest node)) where
      here = if f $ rootLabel node then Just pos else Nothing

inTree :: NodePos -> Tree a -> Bool
pos `inTree` tree = 
  case getTreeNode_ pos tree of
    Just a  -> True
    Nothing -> False

{-|
  Doesn't check bounds, use 'inTree' for that.
-}
getTreeNode :: NodePos -> Tree a -> a
getTreeNode pos tree = 
  case getTreeNode_ pos tree of
    Just a  -> a
    Nothing -> error "getTreeNode: bad position"

getTreeNode_ :: NodePos -> Tree a -> Maybe a
getTreeNode_ (i:is) node = 
  case getIndex (subForest node) i of
    Nothing    -> Nothing
    Just node' -> getTreeNode_ is node'
getTreeNode_ [] node = Just $ rootLabel node

{-|
  Doesn't check bounds, use 'inTree' for that.
-}
updateSubTree :: (Tree a -> Tree a) -> NodePos -> Tree a -> Tree a
updateSubTree f (i:is) node = node {subForest = replaceIndex i x' xs} where
  x' = updateSubTree f is (xs!!i)
  xs = subForest node
updateSubTree f [] node = f node

zipTreeWith :: (a -> b -> c) -> Tree a -> Tree b -> Tree c
zipTreeWith f (Node node1 ts1) (Node node2 ts2) =
  Node (f node1 node2) (zipWith (zipTreeWith f) ts1 ts2)

zipTree :: Tree a -> Tree b -> Tree (a, b)
zipTree = zipTreeWith (,)

{-|
  Numbers all the tree nodes in preorder starting with 1.
-}
numberTree :: Tree a -> Tree Int
numberTree tree = evalState (number tree) 1 where

  number :: Tree a -> State Int (Tree Int)
  number (Node a ts) = do
    i <- get
    modify (+ 1)
    ts' <- mapM number ts
    return (Node i ts')

{-|
  Find the number as would be given by numberTree.
-}
nodePos2Number :: Tree a -> NodePos -> Int
nodePos2Number tree node = 
  let lineNumbers = numberTree tree
  in  getTreeNode node lineNumbers

updateRootLabel :: (a -> a) -> Tree a -> Tree a
updateRootLabel f tree = tree {rootLabel = f $ rootLabel tree}

updateSubForest :: ([Tree a] -> [Tree a]) -> Tree a -> Tree a
updateSubForest f tree = tree {subForest = f $ subForest tree}

-- ==============================
-- * String
-- ==============================

trim :: String -> String
trim = applyTwice (reverse . trim1) where
  trim1 = dropWhile (`elem` delim) 
  delim    = [' ', '\t', '\n', '\r']
  applyTwice f = f . f

{-| 
  Split on space. Understands quoted strings.
-}
tokenize :: String -> [String]
tokenize s = 
  runReadP (many token) s 
  where
  token = do
    skipSpaces
    input <- look
    case input of
      '"' : _ -> lex >>= (\(String s) -> return s)
      ""      -> fail "EOF"
      _       -> munch (not . isSpace)

tryReadP :: ReadP a -> String -> Either String a
tryReadP r s =
  case [ x | (x,"") <- readP_to_S r s] of
    [x] -> Right x
    []	-> Left "no parse"
    _	-> Left "ambiguous parse"

runReadP :: ReadP a -> String -> a
runReadP r s = 
  case tryReadP r s of
    Right x -> x
    Left  e -> error $ "runReadP: " ++ e

